# -*- coding: utf-8 -*-

#DIRDYN.PY
#@author: Thomas Gillis
#
#Created on 3/7/2013
#Last update on 9/7/2013
#

#
# PyRobotran
# Copyright 2013, Universite catholique de Louvain, Belgium
# All rights reserved
# 
# 
# The PyRobotran software has been developed by Thomas Gillis (Louvain School of engineering)
# and Nicolas Docquier (Center for Research in Mechatronics, institute of MEchanics, Materials and 
# Civil engineering) for the Université catholique de Louvain (UCL).
#
# PyRobotran is a free software: you can redistribute it and/or modify it under the terms of 
# the BSD 3-Clause License as published below.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice, 
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice, 
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Université catholique de Louvain nor the names of
#     its contributors may be used to endorse or promote products derived from 
#     this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# For any use in commercial products or questions about licensing, you can contact
#  - the Center for Research in Mechatronics at info@cerem.be
#  - the Louvain Technology Transfer Office of UCL at ltto@uclouvain.be
#


import numpy as np
from scipy.integrate import ode
from scipy.interpolate import interp1d
from numpy.linalg import solve
from numpy import dot
import time
import os
import warnings


class Dirdyn:
    """
    Classe Dirdyn: Direct dynamic resolution  
    
    :param integrator: Name of the numpy integrator to use (supported: vode, lsoda, dopri5; default: vode)
    :param dtmax: Maximum step allowed for the integrator (default: 1e-2)
    :param reltol: Relative tolerance for the integrator (default: 1e-3)
    :param abstol: Absolute tolerance of the integrator (default: 1e-6)
    :param dtinit: First time step of the integrator (default: 1e-3)
    :param framerate: Laps of time for the interpolation in order to smooth the animation file (default: 1e-3)   
    :param result: Result object containing all the results of the integration
    :param mbs: Mbs_Data object containing all data's needed during the integration  
    :param t0: Initial time for the integration (default: 0)    
    :param t1: Final time for the integration (default: 5)
    :param stiff: Boolean to indicate if the problem is stiff or not (vode integrator)
    :param dt: the step size at which we will store the results (all integrators)
    :param displayTime: True if we want to print the time of integration
    
    """
    
    
    def __init__(self,mbs):
        self.dtmax=1e-2
        self.reltol=1e-3
        self.abstol=1e-6
        self.dtinit=1e-3
        self.framerate=0.001
        self.stiff=False
        self.result=Result(mbs)
        self.mbs=mbs
        self.t0=0
        self.t1=5
        self.dt=0.001
        self.displayTime=False
        
        self.integrator='vode'
        self.solver = ode(self.dfdt).set_integrator(self.integrator,first_step=self.dtinit,max_step=self.dtmax,atol=self.abstol,rtol=self.reltol,method='bdf')
        

    def set_output_sensor(self,name,sensorName,vector,index):
        """ This function include the 'name' parameter in the list of all the
        results to save. The information are used to automatically save the data's wanted.
        
        :param name: string containing the name of the output parameter
        :param nameSensor: string containing the name of the sensor which have to measure the parameter
        :param vector: string containing the vector of the sensor wanted: P, OMP,...
        :param index: index of the sensor wanted
        
        """
        self.result.paramToSave[name]=[]
        self.result.infoParam[name]=(sensorName,vector,index)
        self.mbs.Noutput+=1
    
    
    def set_output_other(self,name):
        """ Create a new entry in the dictionnary to save data's. If the name already exists it set the value to zero.
        
        :param name: the name of the parameter
        
        """
        self.mbs.dicoOutOther[name]=0.0
        self.result.paramToSave[name]=[]
        self.mbs.Noutput+=1


    def dfdt(self,t,y):
        """ Compute the derivative of the eqation: M qdd + c = Qq + h ...
        For each time t we update the mbs_data with the new configuration
        and return yd, the derivative vector of y
        
        :param t: the time at which we want to compute the acceleration and velocity
        :param y: a numpy array containing the position and velocity of the independant coordinates [qu qud]
        
        :return yd: a numpy array containing the derivative of y: [qud, qudd]
        
        """
        
        # variable to determine wether post-process data must be computed
        # (Lagrange multpliers, ...)
        # Should be pass in argument
        post_process = True
        
        
        yd=np.array(y) #creation de yd a l'image de y

        #recuperation des indices pour plus de facilite
        ind_qu=self.mbs.qu
        ind_qv=self.mbs.qv
        ind_qc=self.mbs.qc #recuperation des indices commandes   
        ind_quc=self.mbs.quc

        #sauvegarde (position-vitesse) pour calculer la configuration du temps t
        self.mbs.q[ind_qu]=y[1:self.mbs.nqu+1]
        self.mbs.qd[ind_qu]=y[self.mbs.nqu+1:2*self.mbs.nqu+1]
        self.mbs.ux[1:]=y[self.mbs.nqu*2+1:]
        self.mbs.tsim=t

        #==============================================================================#
        #  Calcul des variables commandees,des forces et des variables dependantes     #
        #==============================================================================#
                
        #recuperation des variables commandee. on écrase les donnees de mbs.q
        if self.mbs.nqc !=0:
            self.mbs.fct.user['djoints'](self.mbs,self.mbs.tsim)

        
        #calcul des nouvelles forces exterieures et des links
        self.mbs.calcForces()
        
        #calcul de Bvu:
        #plus besoin de calculer J et h car deja connus par NR!
        if self.mbs.nqv != 0:
            #on calcul les variables dependantes:
                #position:
            self.mbs.close_geo()

                #vitesse:
            Bvu=- solve(self.mbs.J[np.ix_(self.mbs.hu,ind_qv)],self.mbs.J[np.ix_(self.mbs.hu,ind_quc)])
            self.mbs.qd[ind_qv]= dot(Bvu,self.mbs.qd[ind_quc])
                #acceleration: (calcul explicite plus bas avec les accelerations des quc)
            self.mbs.fct.sym['cons_jdqd'](self.mbs.jdqd,self.mbs)
            b=- solve(self.mbs.J[np.ix_(self.mbs.hu,ind_qv)],self.mbs.jdqd[self.mbs.hu])

        #matrice de masse et c
        self.mbs.fct.sym['dirdyna'](self.mbs.M,self.mbs.c,self.mbs,self.mbs.tsim)
        
        
        #=============================================================#
        #         Premier partitionnement de coordonnees              #
        #=============================================================#
               
        try:
            Muu=self.mbs.M[np.ix_(ind_quc,ind_quc)]
            cu=self.mbs.c[ind_quc]
        except IndexError:
            print('ERROR: probleme dans le premier partitionnement, pas d\' indice dependant')
        
        if self.mbs.nqv !=0:
            Muv=self.mbs.M[np.ix_(ind_quc,ind_qv)]
            Mvu=self.mbs.M[np.ix_(ind_qv,ind_quc)]
            Mvv=self.mbs.M[np.ix_(ind_qv,ind_qv)]
            cv=self.mbs.c[ind_qv]
        
        
        #expression des matrices du deuxieme ordre
        A=Muu
        if self.mbs.nqv !=0:
            A += dot(Muv,Bvu) + dot(Bvu.T,Mvu)+dot(dot(Bvu.T,Mvv),Bvu) # = Mruu car pas de driven
        
        F=cu-self.mbs.Qq[ind_quc]
        if self.mbs.nqv != 0:
            F+= dot(Muv+dot(Bvu.T,Mvv),b)+dot(Bvu.T,cv) - dot(Bvu.T,self.mbs.Qq[ind_qv])
        
        #=============================================================#
        #         Second partitionnement de coordonnees               #
        #=============================================================#
        Mruu=A[0:self.mbs.nqu,0:self.mbs.nqu]
        Fru=F[0:self.mbs.nqu]
        B=Fru
        if self.mbs.nqc !=0:
            Mruc=A[0:self.mbs.nqu,self.mbs.nqu:]

            B+= dot(Mruc,self.mbs.qdd[ind_qc])
        
        #=============================================================#
        #         Calcul des accelerations finales et fin             #
        #=============================================================#
        
        #calcul des accelerations:
        self.mbs.qdd[ind_qu]= - solve(Mruu,B)
        if self.mbs.nqv !=0:
            self.mbs.qdd[ind_qv]=dot(Bvu,self.mbs.qdd[ind_quc])+b
        
        #on stocke dans yd et on renvoi le tout
        
        yd[1:self.mbs.nqu+1]=self.mbs.qd[ind_qu]
        yd[self.mbs.nqu+1:2*self.mbs.nqu+1] = self.mbs.qdd[ind_qu]


        #=============================================================#
        #                      Other computations                     #
        #=============================================================#
        
        # Lagrange multipliers associated to constraints
        #=====================================================================#
        if self.mbs.nqv != 0:
            if post_process:
                self.mbs.Lambda = np.zeros(self.mbs.Ncons+1);   # add +1 for keeping indexing from 1 instead of 0
                Jv = self.mbs.J[np.ix_(self.mbs.hu,ind_qv)]
                Fv = cv-self.mbs.Qq[ind_qv]
                lam = solve(Jv.T, dot(Mvu,self.mbs.qdd[ind_quc]) +  dot(Mvv,self.mbs.qdd[ind_qv]) + Fv);
                self.mbs.Lambda[self.mbs.hu] = lam;
               
        
        
        # Forces/torques associated to forced (driven/locked) joints : Q(ind_c)
        #=====================================================================#
        
        if self.mbs.nqc != 0:
            if post_process:
                #self.mbs.M[np.ix_(ind_quc,ind_quc)]
                Mrcc = A[self.mbs.nqu:,self.mbs.nqu:];
                Mrcu = A[self.mbs.nqu:,0:self.mbs.nqu];
                Frc   = F[self.mbs.nqu:];
                self.mbs.Qq[ind_qc] = dot(Mrcu,self.mbs.qdd[ind_qu]) + dot(Mrcc,self.mbs.qdd[ind_qc]) + Frc;
        
        # Compute the user derivatives:  user_deriv
        #=====================================================================#        
        if self.mbs.Nux != 0:
            self.mbs.fct.user['deriv'](self.mbs.ux,yd[self.mbs.nqu*2:],self.mbs,self.mbs.tsim)
            #take care! we should have write 'yd[self.mbs.nquc*2+1:]' but the vectors
            #begin at 0 and for the derivatives function we have to set a zero in front of this vector
            #because there index begin at 1!
            
        return yd
        
        
# = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = =
#  = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = = 
    
    def initTimeIntegrator(self, nsteps=1):
        """ Initialize the time integrator for this direct dynamics:
                - compute the initial state vector (by calling getInitialStateVector)
                - create the integrator object (self.solver)
                - initialze the time integrator object to time self.t0 and to the initial state vector
                - compute the initial acceleration
                - save the initial value of output (q, qd, qdd, userIO, ...)
        
        """
        
        y0 = self.getInitialStateVector()
        
        self.mbs.doneDirdyn=True
        
        #integration
        print('Choice of integrator:{0}'.format(self.integrator))
        
        if self.stiff and self.integrator == 'vode':
            self.solver = ode(self.dfdt).set_integrator(self.integrator,first_step=self.dtinit,max_step=self.dtmax,atol=self.abstol,rtol=self.reltol,method='bdf')

        elif self.integrator == 'dopri5':
            self.solver = ode(self.dfdt).set_integrator(self.integrator,first_step=self.dtinit,max_step=self.dtmax,atol=self.abstol,rtol=self.reltol,nsteps=1)

        else:
            self.solver = ode(self.dfdt).set_integrator(self.integrator,first_step=self.dtinit,max_step=self.dtmax,atol=self.abstol,rtol=self.reltol)
            
        self.solver.set_initial_value(y0, self.t0)
        
        #computing initial acceleration and velocity
        self.dfdt(self.t0,y0) #on calcul acceleration et NR
        
        # Save initial values
        self.result.add(self.mbs)
        if self.mbs.Nux != 0:
            self.result.add_state(self.mbs.ux,self.mbs.tsim)
    
    
    
    def getInitialStateVector(self):
        """ compute and return the initial state vector ([qu, qdu, ux]) """
        
        #loading initial conditions. We take into account the dependent and the driven coordinates
        nbry=2*self.mbs.nqu+self.mbs.Nux
        y0=np.zeros(nbry+1)
        y0[1:self.mbs.nqu+1]=self.mbs.q[self.mbs.qu]
        y0[self.mbs.nqu+1:2*self.mbs.nqu+1]=self.mbs.qd[self.mbs.qu]
        y0[self.mbs.nqu*2+1:]=self.mbs.ux[1:]
        
        return y0
        
        
    def saveStep(self):
        """ usefull function for saving intermediate results at the end of a time step:
               - call dfdt to ensure that acceleration and all output are up to date
               - save tue user state to the result "sub-object"
               - compute and save the requested sensor info
               - compute and save the other requested output
        """
        # rappel de dfdt?? a priori pas necessaire pour lsoda, vode
        self.result.add(self.mbs) #sauvegarde des param par default
        
        #sate            
        if self.mbs.Nux != 0:
            self.result.add_state(self.mbs.ux,self.mbs.tsim)
        
        #parameters
        dico=dict()
        for nameParam in self.result.infoParam.keys():
            info=self.result.infoParam[nameParam]
            sensor=self.mbs.get_sensor(info[0])
            sensor.compute()
            dico[nameParam]=sensor.get_param(info[1],info[2])
        
        for other in self.mbs.dicoOutOther.keys():
            dico[other]=self.mbs.dicoOutOther[other]
        
        self.result.add_other(self.mbs.tsim,dico)                
        
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        
    def run(self):
        """ Compute the direct dynamic solution by integration. The integrator
        is defined in the object construction. The function used to
        compute the derivative is the dfdt(self,t,y) function.
        """
        
        assert self.mbs.donePart == True
        
        print('Starting Computation')
        
        start=time.time()
        
        # initialize the time integration
        self.initTimeIntegrator()
            
            
#       INTEGRATION
#       -----------

        warnings.filterwarnings("ignore")
#        solver._integrator.iwork[2] = -1
        succes=True

        while succes and self.solver.t < self.t1:
            if self.mbs.equil:
                self.solver.integrate(self.t1)
                succes=self.solver.successful()
                
            elif self.integrator == 'vode':
                self.solver.integrate(self.t1, step=True)
                
            elif self.integrator == 'dopri5':
                self.solver.integrate(self.t1, step=True)
                
            else:
                self.solver.integrate(self.solver.t+self.dt)
                succes=self.solver.successful()
            
            if self.displayTime:
                print('time:{}'.format(self.solver.t))
            
            # save the results at the end of the current time step
            self.saveStep()
        
        
        warnings.resetwarnings()
        
        #calcul du temps de calcul
        print('End of the integration after {0: .3} sec'.format(time.time()-start))
        # on met tout ca en array a la fin de l'integration et on sauve
        self.result.toArray()     
        self.toTxt()
                
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    def toTxt(self, name=None):
        """ Write the animation file and save the results.
        
        It will save the results q,qd, qdd and the state by default and it will save the results demanded by 
        the user (function 'set_output').
        
        Take care! The zero in front of each vector is still in the result file.
        So you can directly use the get identity functions!
        
        But before saving the results in the file it interpolate the function
        position-time because the dtime isn't the same in the result object.
        
        Specifying a value for name allows to specify the name of the files in which results and animation
        data are saved. If no name is given, the project name is used.
        
        """
        print('Writting the anim file')
        
        if name is None:
            name = self.mbs.name
        
        filename=os.path.join(self.mbs.prjpath,'animationR',name+'.anim')

        
        #interpolation pour avoir une belle animation
        time=np.arange(self.result.time[0],self.result.time[-1],self.framerate)
        
        # attention il faut transposer les result sinon ca fout le bordel!
        fq=interp1d(self.result.time,self.result.q.T)
        cleanq=fq(time)
        
        tosave=cleanq.T
        tosave[:,0]=time        
        
        # create the animationR folder if it does not exist
        if not os.path.exists(os.path.normpath(os.path.join(filename,'..'))):
            os.makedirs(os.path.normpath(os.path.join(filename,'..')))
        
        np.savetxt(filename,tosave,fmt="%.8f")
        
        print ('writting result files')
        time=np.arange(np.min(self.result.time),np.max(self.result.time),self.framerate)

        resultsRFolder = os.path.join(self.mbs.prjpath,'resultsR')

        # create the results folder if it does not exist
        if not os.path.exists(os.path.normpath(resultsRFolder)):
            os.makedirs(os.path.normpath(resultsRFolder))

        
        #sauvegarde des q
        filename=os.path.join(resultsRFolder,name+'q'+'.res')
        fq=interp1d(self.result.time,self.result.q.T)
        cleanq=fq(time)
        tosave=cleanq.T
        tosave[:,0]=time        
        np.savetxt(filename,tosave)
        
        #sauvegarde des qd
        filename=os.path.join(resultsRFolder,name+'qd'+'.res')
        fq=interp1d(self.result.time,self.result.qd.T)
        cleanq=fq(time)
        tosave=cleanq.T
        tosave[:,0]=time        
        np.savetxt(filename,tosave)
        
        #sauvegarde des qdd
        filename=os.path.join(resultsRFolder,name+'qdd'+'.res')
        fq=interp1d(self.result.time,self.result.qdd.T)
        cleanq=fq(time)
        tosave=cleanq.T
        tosave[:,0]=time        
        np.savetxt(filename,tosave)
        
        
        #state saving
        if self.mbs.Nux != 0:
            time=np.arange(np.min(self.result.timeState),np.max(self.result.timeState),self.framerate)
            filename=os.path.join(resultsRFolder,name+'state'+'.res')
            fq=interp1d(self.result.timeState,self.result.state.T)
            cleanq=fq(time)
            tosave=cleanq.T
            tosave[:,0]=time        
            np.savetxt(filename,tosave)
        
        
        #parameters saving
        if self.mbs.Noutput != 0:
            time=np.arange(np.min(self.result.timeParam),np.max(self.result.timeParam),self.framerate)
            for nameParam in self.result.paramToSave.keys():
                filename=os.path.join(resultsRFolder,name+nameParam+'.res')
                fq=interp1d(self.result.timeParam,self.result.paramToSave[nameParam].T)
                cleanq=fq(time)
                tosave=cleanq.T
                tosave[:,0]=time        
                np.savetxt(filename,tosave)
        


    def setTspan(self,t0,t1):
        """ Set the begining and the final time for the integration

        :param t0: stating time of the integration
        :param t1: end time of the simulation

        """
        self.t0=float(t0)
        self.t1=float(t1)
    
    
    def setParam(self,**dico):
        """ Set the parameters of the integrator used for the integration
        
        :param dico: the dictionnary containing the named arguments of the function.
        
        The recognised arguments are:
            * integrator: the integrator to use to solve de ODE's
            * abstol: the absolute tolerance of the integrator
            * reltol: the relative tolerance of the integrator
            * dtmax: the maximum time step of the integrator
            * dtinit: the first time step of the integrator
            * stiff: Boolean to indicate if the problem is stiff or not (vode integrator)
            * dt: The stepsize for the integrators. We will stop the integrator each dt seconds and store the result

        """
        for param in dico.keys():
            if param == 'integrator':
                self.integrator= dico[param]
            elif param == 'dtmax':
                self.dtmax=dico[param]
            elif param == 'reltol':
                self.reltol=dico[param]
            elif param == 'abstol':
                self.abstol=dico[param]
            elif param == 'dtinit':
                self.dtinit=dico[param]
            elif param == 'stiff':
                self.stiff=dico[param]
            elif param == 'dt':
                self.dt=dico[param]
            else:
                print('unknown parameter')


class Result:
    """Class result used to store the position, velocity and acceleration of the
    multibody system at each time step of the integrator.
    The values are store under list structure. At the end of the computation we have to 
    call the 'toArray' function to set them as numpy arrays.
    
    :param time: time at step
    :param q: position at each time step
    :param qd: velocity at each time step
    :param qdd: acceleration at each time step
    :param paramToSave: dictionnary with key the name of the parameter to save and the values are lists. This lists save the value of the parameter wanted at each time step
    :param timeParam: list containing the time at which the parameters have been saved
    :param state: state vector at each time step
    :param timeState: list containing the time at which the states have been saved
    :param mbs: the multibody data structure
    """
    def __init__(self,mbs):        
        self.q=[]
        self.qd=[]
        self.qdd=[]
        self.Qq=[]
        self.Lambda=[]
        self.time=[]
        self.paramToSave={}
        self.timeParam=[]
        self.infoParam={}
        self.state=[]
        self.timeState=[]
        self.mbs=mbs

    def add(self,mbs): # q,qd et qdd des np.arays
        """ Add the results q, qd and qdd to the different lists by appending the 
        value. It also save the time. This function is automaticaly called by the program
        to save the position-velocity and acceleration.
        
        :param time: time at which the q, qd and qdd have been computed
        :param q: numpy array with the position values at time 'time'
        :param qd: numpy array with the velocity values at time 'time'
        :param qdd: numpy array with the acceleration values at time 'time'
        
        """
        self.time.append([mbs.tsim])
        # on append q en le passant dans la liste grace a la fonction .tolist()
        self.q.append(mbs.q.tolist())
        self.qd.append(mbs.qd.tolist())
        self.qdd.append(mbs.qdd.tolist())
        
        self.Qq.append(mbs.Qq.tolist())
        
        if mbs.nqv!=0:
            self.Lambda.append(mbs.Lambda)

    def add_state(self,ux,time):
        self.state.append(ux.tolist())
        self.timeState.append([time])

        

    def add_other(self,tsim,valueToStore):
        """ Add the results set in the dictionnary valueToStore to the lists.
        This function has to be called by the user to store what he want to save.
        The user don't have to give a value for all the parameter he entered in the 
        'set_output' function. If a couple (parameter,value) is not given. The function
        will set a 0 at time 'time'        
        
        :param time: time at which the values have been computed
        :param valueToStore: dictionnary. The keys are the name of the parameter, the same name as given to the 'set_output' function. The value is the value of the parameter at time 'time'.
        
        """
        
        self.timeParam.append([tsim])
        for key in self.paramToSave.keys():
            if key in valueToStore:
                insert=[valueToStore[key]]
                insert.insert(0,0.0)
                self.paramToSave[key].append(insert)
            else:
                self.paramToSave[key].append([0.0,0.0])
    
    def toArray(self):
        """ This function convert the lists q, qd, qdd, time, timeParam and
        paramToSave into numpy arrays. The program only store numpy arrays in
        the animation file.
        """
        self.q=np.array(self.q)
        self.qd=np.array(self.qd)
        self.qdd=np.array(self.qdd)
        
        self.Qq=np.array(self.Qq)
        self.Lambda=np.array(self.Lambda)

        #flating the array to get an array of one dimension
        self.time=np.array(self.time).flat[:]
        

        #parameters
        if self.mbs.Noutput != 0:
            self.timeParam=np.array(self.timeParam).flat[:]
            for key in self.paramToSave.keys():
                self.paramToSave[key]=np.array(self.paramToSave[key])
            
        #state
        if self.mbs.Nux != 0:
            self.state=np.array(self.state)
            self.timeState=np.array(self.timeState).flat[:]
            












