# -*- coding: utf-8 -*-

# MBS_DATA.PY
# @author: Thomas Gillis
#
# Created on 1/7/2013
# Last update on 07/10/2013
#

#
# PyRobotran
# Copyright 2013, Universite catholique de Louvain, Belgium
# All rights reserved
# 
# 
# The PyRobotran software has been developed by Thomas Gillis (Louvain School of engineering)
# and Nicolas Docquier (Center for Research in Mechatronics, institute of MEchanics, Materials and 
# Civil engineering) for the Université catholique de Louvain (UCL).
#
# PyRobotran is a free software: you can redistribute it and/or modify it under the terms of 
# the BSD 3-Clause License as published below.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice, 
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice, 
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Université catholique de Louvain nor the names of
#     its contributors may be used to endorse or promote products derived from 
#     this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# For any use in commercial products or questions about licensing, you can contact
#  - the Center for Research in Mechatronics at info@cerem.be
#  - the Louvain Technology Transfer Office of UCL at ltto@uclouvain.be
#


import numpy as np
import collections as col
from numpy.linalg import solve,norm
import sys
from xml2sys import *


class Data:
    """ class Mbs_Data: contain all the data's from the multibody system
    
    :param name: the name of the multibody system
    :param prjpath: the absolute path to the main project folder. For example: '/home/Dupont/Desktop/Robotran/myProject'
    :param sympath: the absolute path to the symbolic folder. For example: '/home/Dupont/Desktop/Robotran/myProject/symbolicR'
    :param userpath: the absolute path to the user fonction folder. For example: '/home/Dupont/Desktop/Robotran/myProject/userfctR'
    :param fct: the function object containing all the references to the symbolic and user functions
    :param g: the numpy array containing the gravity vector
    :param dpt: numpy array with the anchor points
    :param npt: number anchor points
    :param l: numpy array with the center of mass
    :param m: numpy array with the mass vector
    :param In: numpy array with the inertia matrix
    :param dicoBody: dictionnary with the Body name and the index of the body in the 'q' vector
    :param dicoJoint: dictionnary with the joint name and the index of the joint in the 'q' vector
    :param njoint: number of joints
    :param lrod: numpy array with the length of each rod
    :param Ncons: number of algebraic constraints
    :param Nurscons: number of user constraints
    :param usrModel: user model object
    :param dicoState: Dictionnary containing the state name and the index of it
    :param Nstate: Number of states
    :param ux: numpy array containing the values of the states for the 'tsim' time
    :param donePart: True if the partionning has been done, false if not
    :param doneDirdyn: True if the direct dynamics has already be run, False if not
    :param doneInvdyn: True if the inverse dynamics has already be run, False if not
    :param doneEqui: True if the equilibrium has been computed, False if not
    :param q: the numpy array containing the position at time tsim
    :param qd: the numpy array containing the velocity at time tsim
    :param qdd: the numpy array containing the acceleration at time tsim
    :param q0: the initial position
    :param qd0: the initial velocity
    :param qdd0: the initial acceleration
    :param qu: list with the index of the independant coordinates
    :param qv: list with the index of the dependant coordinates
    :param qc: list with the index of the forces coordinates
    :param quc: list with the index of the independant and forces coordinates
    :param qa: list with the index of the motorised coordinates (NOT IN USE)
    :param qlocked: list with the index of the locked coordinates (NOT IN USE)
    :param nqlocked: number of locked coordinates (NOT IN USE)
    :param nqv: number of dependant coordinates
    :param nqu: number of independant coordinates
    :param nqc: number of driven coordinates
    :param nqa: number of motorised coordinates (NOT IN USE)
    :param nquc: number of indenpendant and dependant coordinates
    :param Qq: numpy array with the forces in the coordinates
    :param Lambda: numpy array with the lagrange multiplier (NOT IN USE)
    :param hu: the index of the dependant constraints
    :param hv: the index of the redundant constraints
    :param frc: numpy array with the forces component on each joint
    :param trq: numpy array with the torque component on each jont
    :param tsim: the simulation time
    :param Nlink: the number of link
    :param Nlink3D: the number of link3D
    :param Z: numpy array with the distance between the two anchor points for each link
    :param Zd: numpy array with the velocity between the two anchor points for each link
    :param Fl: numpy array with the force in each link
    :param dicoLink: dictionnary containing the link and the index corresponding two this link
    :param Nsensor: the number of sensor
    :param dicoSensor: the dictionnary containing the sensors objects
    :param Nxfrc: the number of external forces
    :param xfidpt: list with the index of the application point of forces.
    :param dicoForce: dictionnary containing the force index
    :param Nwheel: the number of wheel (NOT IN USE)
    :param rnom: the nominal radius
    :param c: the c vector
    :param M: the mass vector
    :param h: the constraints vector
    :param J: the jacobian matrix
    :param jdqd: the derivative of the jacobian
    :param Noutput: the number of output wanted by the user
    :param equil: is True if we are computing the Equilibrium, False if not
    :param NOutOther: the number of output which isn't sensors
    :param dicoOutOther: the dictionnary of outputs (no sensors output) containing the name and the value at time tsim
    
    """
    
    def __init__(self,filename):
        self.filename = filename
        
        
        self.g=np.zeros(3)
        
        self.dpt=np.zeros(3) #tableau 3*N qui contient les points d'ancrages
        self.npt=0 #nombre de points d'ancrages
        
        self.l=[] #tableau qui contient les centre de masse 3*N
        self.m=[] #tableau qui contient les masses 1*N
        self.In=[] #tableau 6*N qui contient les inerties
        
        self.dicoBody=col.OrderedDict() # permet de stocker les 'id' des corps
#        self.inbody=[]# dictionnaire qui a pour cle un indice de corps et comme
#            # valeur un indice d'articulation (et donc de coord. gen.)
#            #non utilisé! 
        self.dicoJoint=col.OrderedDict() #dictionnaire faisant correspondre un joint donne (cle)
            # avec un indice de joint qui permet de le retrouver dans les q,qd et qdd
        self.njoint=0 #nombre de JOINTS!
        
        self.lrod=[0] #longueur de chaque rod.
        #On commence par une liste et ce sera transforme en tableau a la fin du chargement dans le fonction get_data
        #On met aussi deja un zero dedans car le futur tableau commence a l'indice 1 et pas a l'indice 0
        self.Ncons=0 #nombre de contraintes algebrique
        self.Nuserc=0 #nombre de contrainte utilisateur!
        
        self.usrModel=None #Dictinnaires des user_models
        self.dicoState=col.OrderedDict()
        self.ux=[]
        self.Nux=0
        
        self.donePart=False #indique si le partionnement a ete fait
        self.doneDirdyn=False #indique si on a fait tourner dirdyn
        self.doneInvdyn=False #indique si on a fait tourner invdyn
        self.doneEqui=False #indique si on a fait tourner equilibrium
        
        self.q=[] #position
        self.qd=[]  #vitesse 
        self.qdd=[] #acceleration
        self.q0=[]  #position initiale
        self.qd0=[] #vitesse initiale
        self.qdd0=[] #acceleration initiale
        self.qmin=0 #min of the gen coord.
        self.qmax=0  #max of the gen coord.
    
    
        self.qu=[] #liste avec les indices des articulations independantes
        self.qv=[] #liste avec les indices des articulations dependante
        self.qdriven=[] #liste avec les indices des articulations driven
        self.qlocked=[] #liste avec les indices des articulations locked
        self.qc=[] #liste avec les indices des articulations driven+locked
        self.quc=[]
        self.qa=[] #liste avec les indices des articulations motorisee
        self.nqv=0
        self.nqu=0
        self.nqc=0
        self.nqdriven=0
        self.nqlocked=0 
        self.nqa=0
        self.nquc=0
        
        self.Qq=[] #tableau 1*N des forces articulaires
        self.Lambda=[] #tableau avec les mutliplicateurs de Lagrange
        self.hu=[]  #contrainte dependantes
        self.hv=[] #contraintes independantes
        self.frc=[] #tableau 3*N: forces sur le centre de masse de chaque corps
        self.trq=[] #idem pour les couples
        self.tsim=0
        
        self.Nlink=0 #nombre de forces de liens
        self.Z=[] #tableau avec les distances entre les deux points
        self.Zd=[] #tableau avec les vitesses entre les deux points
        self.Fl=[] #tableau avec les forces de lien 
        self.dicoLink=col.OrderedDict() #dictinonaire avec les link. Permet de retrouver les id des link

        self.Nlink3D=0 #nombre de forces de liens
        self.dicoLink3D=col.OrderedDict() #dictinonaire avec les link. Permet de retrouver les id des link
        
        self.Nsensor=0 #nombre des senseurs
        self.dicoSensor=col.OrderedDict()
        
        self.Nxfrc=0 #nombre de points avec une force externe
        self.xfidpt=[0.0]  #tableau des indices des points des forces d'applications
        self.dicoForce=col.OrderedDict()
        
        
        self.Nwheel=0 #nombre de roue sur le modele
        self.rnom=[] #tableau avec les rayons des roues
        
        #saving computing data's
        self.c=[]
        self.M=[]
        self.h=[]
        self.J=[]
        self.jdqd=[]
        
        self.equil=False
        self.Noutput=0
        
        self.dicoOutOther={}
    
    
    def add_output_other(self,name,value):
        """ Function to store the value at the name 'name'. It first check is the name is known and then it replace the
        value already existing.
        
        :param name: the name of the parameter to save. The name has to be previously entered with the dirdyn function 'set_output_other'
        :param value: the value to store
        
        """
        
        if name in self.dicoOutOther.keys():
            self.dicoOutOther[name]=value
        else:
            print('unkown name!')
            
        
    def reset_qu_qv(self, qu, qv):
        """ Reset the qu and qv vector to the list of indices given in 
            argument.
            The indices given in qu and qv must already be either dependent
            or independent (but not driven or forced).
            This method is intented to be used for the coordinate partitioning
            qu and qv are not sorted
        """
        # first check that indices of qu and qv are already in qu or in qv
        quv_new = (qu+qv)
        quv_new.sort()
        quv_old = (self.qu+self.qv)
        quv_old.sort()
        assert quv_new==quv_old
        
        
        # update the array of the C strcture
        self.set_coord_nature_C(qu, qv, self.qdriven, self.qlocked)        
               

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        
    def reset_qu_qv(self, qu, qv):
        """ Reset the qu and qv vector to the list of indices given in 
            argument.
            The indices given in qu and qv must already be either dependent
            or independent (but not driven or forced).
            This method is intented to be used for the coordinate partitioning
            qu and qv are not sorted
        """
        # first check that indices of qu and qv are already in qu or in qv
        quv_new = (qu+qv)
        quv_new.sort()
        quv_old = (self.qu+self.qv)
        quv_old.sort()
        assert quv_new==quv_old
        
        self.qu = list(qu)
        self.qv = list(qv)    
        
        self.nqu = len(qu)
        self.nqv = len(qv)     
        
        self.__update_quc()   
    
    def set_qu(self, ids):
        """Set the index of coordinates in the ids list as independant
        
        :param ids: list of coordinates to set to independant
        
        """
        self.set_nature_toListId(ids, 1)

    def set_qv(self, ids):
        """Set the index of coordinates in the ids list as dependant
        
        :param ids: list of coordinates to set to dependant
        
        """
        self.set_nature_toListId(ids, 2)


    def set_qdriven(self, ids):
        """Set the index of coordinates in the ids list as driven
        
        :param ids: list of coordinates to set to driven
        
        """
        self.set_nature_toListId(ids, 3)


    def set_qlocked(self, ids):
        """Set the index of coordinates in the ids list as locked
        
        :param ids: list of coordinates to set to locked
        
        """
        self.set_nature_toListId(ids, 4)


    def set_nature_toListId(self, ids, toListId):
        """ Add the coord index in ids to the list corresponding to the
            toListId. toListId is defined as follow:
            - if toListId=1: add to qu (independent) and remove from the other lists
            - if toListId=2: add to qv (dependent) and remove from the other lists
            - if toListId=3: add to qdriven (driven) and remove from the other lists
            - if toListId=4: add to qlocked (locked) and remove from the other lists                                       
        """
        # get a local copy of the list of indices
        qu, qv, qdriven, qlocked = self.qu, self.qv, self.qdriven, self.qlocked
        natureLists = [qu, qv, qdriven, qlocked]
        
        # get the ids of the from lists 
        toList = natureLists[toListId-1]
        natureLists.remove(toList)
        
        # update the local copy
        hasModified = self.set_nature(ids, toList, natureLists)
        
                
        
        self.nqu = len(qu)
        self.nqv = len(qv)
        self.nqdriven = len(qdriven)
        self.nqlocked = len(qlocked)
        self.__update_quc()

        
    def set_nature(self, ids, toList, fromLists):
        """ Add the coord index in ids to the toList list and remove
            them from the fromLists list
            ids: a list of coord index
            toLists: the list of index in which to add element of ids
            fromLists: a list of lists (!) of index from which to remove the index
            
            return True if a modification has been done (trensfer of at least one index
                        from one list to another), False otherwise
        
        :param list: list of coordinates to set to independant
        
        """
         
        # flag to check whether a modification has been done
        hasModif = False
        
        # update the given list
        for index in ids:
            if index not in toList:
                # add it to the independent joint list
                toList.append(index)
                
                #remove from previous list
                self.removeIdFromLists(index, fromLists)
                
                hasModif = True
        
        if hasModif:
            # sort indices
            toList.sort()
        
        return hasModif    

    def removeIdFromLists(self, ind, lists):
        """ Utilities function for removing the index from a
            set of list of index.
            lists must be a list of list of index.
            The function search for ind in the list and remove it if it 
            find it.
            The function return after having found the first occurence
        """
        for lis in lists:
            if ind in lis:
                lis.remove(ind)
                return    
                
                            
    def __update_quc(self):
        # update qc list
        self.qc = self.qdriven + self.qlocked
        self.qc.sort()
        self.nqc = self.nqdriven + self.nqlocked

        #update quc list
        self.quc = self.qu + self.qc
        self.nquc = self.nqu + self.nqc

    def reset_hu_hv(self, hu, hv):
        huv = hu+hv
        huv.sort()
        assert huv==range(1, self.Ncons+1)
        
        self.hu = list(hu)
        self.hv = list(hv)
        self.nhu = len(hu)
        self.nhv = len(hv)
        
        
    
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
            
        
    def get_body_id(self,name):
        """Return the index of the body named 'name'
        
        :param name: string containing the name of the body wanted
        
        :return index: the index of the body
        
        """
        try:
            bodyid=self.dicoBody[name] #recherche dans le dictionnaire
        except KeyError:
            print('Unable to find the body:{0}'.format(name))
            bodyid=0
        return bodyid
    
    def get_joint_id(self,name):
        """Return the index of the joint named 'name'
        
        :param name: string containing the name of the joint wanted
        
        :return index: the index of the joint
        
        """
        try:
            jointid=self.dicoJoint[name] #recherche dans le dictionnaire
        except KeyError:
            print('Unable to find the joint:{0}'.format(name))
            jointid=0
        return jointid
    
    def get_link_id(self,name):
        """Return the index of the link named 'name'
        
        :param name: string containing the name of the link wanted
        
        :return index: the index of the link
        
        """
        try:
            linkid=self.dicoLink[name] #recherche dans le dictionnaire
        except KeyError:
            print('Unable to find the link:{0}'.format(name))
            linkid=0
        return linkid
        
    def get_force_id(self,name):
        """Return the index of the ext forces named 'name'
        
        :param name: string containing the name of the external force wanted
        
        :return index: the index of the force
        
        """
        try:
            forceid=self.dicoForce[name].number
        except KeyError:
            print ('Unable to find the force:{0}'.format(name))
            forceid=0
        return forceid
        
    def get_state_id(self,name):
        """Return the index of the state named 'name'
        
        :param name: string containing the name of the state wanted
        
        :return index: the index of the state
        
        """
        try:
            stateid=self.dicoState[name]
        except KeyError:
            print ('Unable to find the state:{0}'.format(name))
            stateid=0
        return stateid
        
    def get_user_model(self,nameModel,nameParam):
        """ Return the value of parameter 'nameParam' of the user model 'nameModel'
        
        :param nameModel: string containing the name of the user model
        :param nameParam: string containing the name of the parameter
        
        :return value: the value of the parameter
        
        """
        param=self.usrModel.dico[nameModel].dicoParam[nameParam]
        if param.type == 'scalar':     
            return param.value[0]
        else:
            return param.value
    

    
    def get_sensor(self,name):
        """return the sensor which has the name 'name'
        
        :param name: string containing the name of the sensor
        
        :return sensor: the sensor object
        
        """
        try:
            return self.dicoSensor[name]
        except KeyError:
            print ('Impssible to find the sensor: {0}'.format(name))
            return None

    def get_F_sensor(self,name):
        """return the force-sensor which has the name 'name'
           this sensor allow to compute the kinematics of the
           sensor associated to an external force

        :param name: string containing the name of the sensor
        
        :return sensor: the sensor object
        
        """
        try:
            return self.dicoForce[name]
        except KeyError:
            print ('Impssible to find the FSensor: {0}'.format(name))
            return None

    
    def load(self):
        """ load the mbs_data structure: It creates a system stucture and transform this stucture
        into the mbs_data structure by calling the 'get_data' function
        """
        f = self.filename
        
        # try to automaticaly define the data file
        # 1st, if the function is called from 'workR' directory --> search in the dataR directory (the most common case)
        if (os.path.basename(os.getcwd())== 'workR' and os.path.isfile(os.path.join(os.getcwd(),'..','dataR',os.path.basename(f))) ):
            self.prjpath = os.path.realpath(os.path.join(os.getcwd(),'..'))
            self.mbsfile = os.path.realpath(os.path.join(os.getcwd(),'..','dataR',f))
            
        # 2nd, if the function is called from the project directory --> search in the dataR directory
        elif os.path.isfile(os.path.realpath(os.path.join(os.getcwd(),'dataR',f))) :
#            self.prjpath = os.getcwd()
            self.mbsfile = os.path.realpath(os.path.join(os.getcwd(),'dataR',f))
            self.prjpath = os.path.realpath(os.path.join(self.mbsfile,'../..'))
            
        # 3rd, assume that the mbs file is in the current directory --> included in the 4th case
#        elif (os.path.isfile(os.path.abspath(f))):
#            self.mbsfile = os.path.abspath(f)
#            if os.path.basename(os.path.join(self.mbsfile,'..'))=='dataR':
#                self.prjpath = os.path.join(self.mbsfile,'..')
#            else:
#                raise Exception('Your mbs file is not in a standard Robotran project folder')
            
        # 4th, assume that the script is calle from anywhere
        elif os.path.isfile(os.path.abspath(f)):
            self.mbsfile = os.path.abspath(f)
            if os.path.basename(os.path.realpath(os.path.join(self.mbsfile,'..')))=='dataR':
                self.prjpath = os.path.realpath(os.path.join(self.mbsfile,'../..'))
            else:
                raise Exception('Your mbs file is not in a standard Robotran project folder')
            
        else:
            raise Exception('Impossible to load the file: '+f)
        
        print("prjpath: "+self.prjpath)
        print("mbsfile: "+self.mbsfile)
        
        # other project subfolders
        self.sympath=os.path.join(self.prjpath,'symbolicR')
        self.userpath=os.path.join(self.prjpath,'userfctR')
        
        mysys=System(self) #lecture du fichier XML
        self.sys2data(mysys) #transfert vers mbs
        
        
    def sys2data(self,sys):
        """ Transform the 'sys' system into the mbs structure and fill the field of the mbs structure

        :param sys: the system structure of the multibody system
        
        """
    
        # getting the project name: 
        self.name = sys.name
        
        # reference to the user and symbolic functions.
        self.fct=Function(self)
        
        
        #======================================================================
        # GRAVITE
        #======================================================================
        self.g=np.zeros(3+1)
        self.g[1:]=sys.base.gravity
        
        
        #======================================================================
        # POINTS D'ANCRAGES (+ comptage du nombre de joints)
        #======================================================================
        #nombre de points d'ancrage: que pour la base ou les body!
        self.npt=len(sys.base.point)
        for key in sys.body.keys():
            self.npt+= len(sys.body[key].point)
        
        #position des points d'ancrages
        self.dpt=np.zeros((3+1,self.npt+1))
        index=1
        indexForce=1
        indexSensor=1
            #base
        for key in sys.base.point.keys():
            self.dpt[1:,index]=sys.base.point[key].value
            if sys.base.point[key].force:
                fSensor=Sensor(self)
                fSensor.number=indexForce
                self.dicoForce[sys.base.point[key].forceName]=fSensor
                self.xfidpt.append(index)
                indexForce+=1
            if sys.base.point[key].sensor:
                sensor=Sensor(self)
                sensor.number=indexSensor
                self.dicoSensor[sys.base.point[key].sensorName]=sensor
                indexSensor+=1
            index+=1
        
            
            #body
        for key in sys.body.keys():
            #rien a voir avec les points d'ancrages, mais on en profite pour avoir
            #   le nombre de joints
            self.njoint += len(sys.body[key].joint)
            #fin du rien a voir avec les points d'ancrages
            
            for keyJ in sys.body[key].point.keys():
                self.dpt[1:,index]=sys.body[key].point[keyJ].value
                if sys.body[key].point[keyJ].force:
                    fSensor=Sensor(self)
                    fSensor.number=indexForce
                    self.dicoForce[sys.body[key].point[keyJ].forceName]=fSensor
                    self.xfidpt.append(index)
                    indexForce+=1
                if sys.body[key].point[keyJ].sensor:
                    sensor=Sensor(self) #on cree un nouveau sensor
                    sensor.number=indexSensor
                    self.dicoSensor[sys.body[key].point[keyJ].sensorName]=sensor
                    indexSensor+=1
                index+=1
        
        #======================================================================
        # NOMBRES DE SENSEURS
        #======================================================================
        #calcul du nombre de senseur:
        self.Nsensor=indexSensor-1
        self.Nxfrc=indexForce-1
        
        #======================================================================
        #JOINTS
        #======================================================================
        #joint: détermination des conditions initiales, dicoJoint, qu,qv et qc et
            #il y a autant de coordonnees generalisees que de joint:
        self.q0=np.zeros(self.njoint+1)
        self.qd0=np.zeros(self.njoint+1)
        self.qdd0=np.zeros(self.njoint+1)    
        
        index=1
        for key in sys.body.keys():
            nbr_joint_for_that_body=len(sys.body[key].joint)
            for keyJ in sys.body[key].joint.keys():
                #valeurs initiales
                self.q0[index]=sys.body[key].joint[keyJ].initialval[0]
                self.qd0[index]=sys.body[key].joint[keyJ].initialval[1]
                self.qdd0[index]=sys.body[key].joint[keyJ].initialval[2]
                
                #ajout du nom du joint dans le dico
                self.dicoJoint[keyJ]=index
                
                #ajout si il est dependant, independant ou driven
                if sys.body[key].joint[keyJ].nature == 'dependent':
                    self.qv.append(index)
                elif sys.body[key].joint[keyJ].nature == 'independent':
                    self.qu.append(index)
                elif sys.body[key].joint[keyJ].nature == 'driven':
                    self.qdriven.append(index)
                elif sys.body[key].joint[keyJ].nature == 'locked':
                    self.qlocked.append(index)
                else:
                    print('ERROR: No (or wrong) kind of joint!!!:{0}, type: {1}'.format(sys.body[key].joint[keyJ].nature,type(sys.body[key].joint[keyJ].nature)))
                    
                #ajout de la parentalité du joint
                if sys.body[key].joint[keyJ].nbr == nbr_joint_for_that_body:
                    self.dicoBody[key]=index
#                    self.inbody.append(index)
                index+=1  
                      
        
        #======================================================================
        # BODY
        #======================================================================
        self.m=np.zeros(self.njoint+1)
        self.l=np.zeros((3+1,self.njoint+1))
        self.In=np.zeros((9+1,self.njoint+1))
        indexbody=0
        for key in sys.body.keys():
#            index=self.inbody[indexbody]
            index=self.get_body_id(key)
            self.m[index]=sys.body[key].mass
            self.l[1,index]=sys.body[key].com[0]
            self.l[2,index]=sys.body[key].com[1]
            self.l[3,index]=sys.body[key].com[2]
            self.In[1,index]=sys.body[key].inertia[0]
            self.In[2,index]=sys.body[key].inertia[1]
            self.In[3,index]=sys.body[key].inertia[2]
            self.In[4,index]=sys.body[key].inertia[1]
            self.In[5,index]=sys.body[key].inertia[3]
            self.In[6,index]=sys.body[key].inertia[4]
            self.In[7,index]=sys.body[key].inertia[2]
            self.In[8,index]=sys.body[key].inertia[4]
            self.In[9,index]=sys.body[key].inertia[5]
            
            indexbody+=1
        
        
        #======================================================================
        # BRICOLES...
        #======================================================================
        # mbs.data.nqu,nqv et nqc
        self.__update_quc()
        
        
        self.nqu=len(self.qu)
        self.nqv=len(self.qv)
        self.nqdriven=len(self.qdriven)
        self.nqlocked=len(self.qlocked)
        self.nqc=len(self.qc)
        
        #on met a jour la stucture mbs_data pour pouvoir demarer avec les CI:
        self.q=np.array(self.q0)
        self.qd=np.array(self.qd0)
        self.qdd=np.array(self.qdd0)
        
        # on met les frc et les trq a 0
        self.frc=np.zeros((3+1,self.njoint+1))
        self.trq=np.zeros((3+1,self.njoint+1))
        
        #on met Qq a zero
        self.Qq=np.zeros(self.njoint+1)
        
        #on remplis les contraintes: lrod et Ncons
        for key in sys.cut.keys():
            self.Ncons+= sys.cut[key].nbrAlgCons
            if sys.cut[key].type == 'rod':
                self.lrod.append(sys.cut[key].length)
                
        self.lrod=np.array(self.lrod, dtype=np.float64)  # add dtype=np.float64 for enforcing an array of double
        
        #usrModel:
        self.usrModel=sys.userModel
        
        # user io (user variables)
        self.usrIO=sys.userIO
        
        #reprise des etats:
        self.Nux =0
        self.ux=[0]
        for model in self.usrModel.dico.keys():
            for param in self.usrModel.dico[model].dicoParam.keys():
                if self.usrModel.dico[model].dicoParam[param].type == 'state':
                    self.ux.append(self.usrModel.dico[model].dicoParam[param].value)
                    self.Nux+=len(self.usrModel.dico[model].dicoParam[param].value)
                    self.dicoState[param]=self.Nux
        
        self.ux=np.array(self.ux)
        
        
        #links
        index=1
        for key in sys.link.keys():
            self.Nlink+=1
            self.dicoLink[sys.link[key].name]=index

        self.Z=np.zeros(self.Nlink+1)
        self.Zd=np.zeros(self.Nlink+1)
        self.Fl=np.zeros(self.Nlink+1)

        #link3D
        index=1
        for key in sys.link3D.keys():
            self.Nlink3D+=1
            self.dicoLink3D[sys.link3D[key].name]=index

        
        
        #======================================================================
        # Initialisation computing data's
        #======================================================================       
        
        self.c=np.zeros(self.njoint+1)
        self.M=np.zeros((self.njoint+1,self.njoint+1))
        self.Ncons+=self.Nuserc
#        print self.Ncons
#        print self.Nuserc
#        print ncons
        self.h=np.zeros(self.Ncons+1)
        self.J=np.zeros((self.Ncons+1,self.njoint+1))
        self.jdqd=np.zeros(self.Ncons+1)
        
        
    def close_geo(self):
        """ Close the cinematic loop thanks to the Newton-Raphson algorithm.
        It compute the dependant coordinates from the value of the independant one.
        """
        ind_qv=self.qv
        tol=1e-9
        err=tol+1
        itermax=20
        n=1
        
        while n<itermax and err > tol:         
            vold=self.q[ind_qv]
            
            self.calc_hJac()
            self.q[ind_qv]= self.q[ind_qv]-solve(self.J[np.ix_(self.hu,ind_qv)],self.h[self.hu])
            err=norm(vold-self.q[ind_qv])
            n+=1
        
        if n==itermax:
            print('Newton-Raphson algorithm has not converged at time {}'.format(self.tsim))
            return -1
        else:
            return 1
        
        
    def calc_hJac(self):
        """ Compute the h and Jacobian matrix of the multibody system. It use the symbolic
        function and the user function to compute the complete h and J matrix.
        """
        self.J.fill(0)
        self.h.fill(0)
        
        self.fct.sym['cons_hJ'](self.h,self.J,self)
        husr=np.zeros(self.Nuserc+1)
        Jusr=np.zeros((self.Nuserc+1,self.njoint+1))
        self.fct.user['cons_hJ'](husr,Jusr,self,self.tsim)
        self.J[self.Ncons-self.Nuserc+1:][:]= Jusr[1:][:]
        self.h[self.Ncons-self.Nuserc+1:]=husr[1:]        
        
    
    def calcForces(self):
        """ Compute forces in the system. It resets the forces and torques at 0
        an then compute self.mbs.frc and self.mbs.trq by applying the link and the extForces
        symbolic functions.
        """
        #mise a zero des forces
        self.frc=np.zeros((3+1,self.njoint+1))
        self.trq=np.zeros((3+1,self.njoint+1))
        
        #forces de lien
        if self.Nlink !=0:
            self.fct.sym['link'](self.frc,self.trq,self.Fl,self.Z,self.Zd,self,self.tsim)

        #forces de lien3D
        if self.Nlink3D !=0:
            self.fct.sym['link3D'](self.frc,self.trq,self,self.tsim)

        
        #forces externes
        if self.Nxfrc != 0:
            self.fct.sym['extforces'](self.frc,self.trq,self,self.tsim)
            
        #        userJF.user_JointForces(self.mbs,self.mbs.tsim)
        try:
            self.fct.user['fjoint'](self,self.tsim)
        except TypeError:
            self.Qq=np.zeros(self.Qq.shape)
            
            
        
class Sensor:
    """ Class Sensor: structure to store a sensor
    
    :param number: Sensor index. This index begin to 1
    :param J: the jacobian of the sensor
    :param P: the position of the sensor
    :param R: rotation matrix of the sensor
    :param V: velocity of the vector
    :param OM: 
    :param A: acceleration of the vector
    :param OMP: 
    :param computeFun: symbolic function to compute sensors
    :param mbs: the mbs_data structure of the multibody system
    
    """
    def __init__(self,mbs):
        self.number=0
        self.J=np.zeros((6+1,mbs.njoint+1))
        self.P=np.zeros(3+1)
        self.R=np.zeros((3+1,3+1))
        self.V=np.zeros(3+1)
        self.OM=np.zeros(3+1)
        self.A=np.zeros(3+1)
        self.OMP=np.zeros(3+1)
        self.computeFun=mbs.fct.sym['sensor']
        self.mbs=mbs
        
    def compute(self):
        """ Compute the data's of the sensor by calling the symbolic function
        """
        try:
            self.computeFun(self,self.mbs,self.number)
        except TypeError:
            raise Exception('No sensor! Error Sensor.compute')
    
    def get_param(self,name,index):
        """ Return the value stored in the vector 'name' at index 'index' of the
        current sensor

        :param name: name of the parameter wanted: P, J, R, V, OM, A, OMP
        :param index: index of the vector wanted. If matrix this can be a tuple
        
        :return value: the value of the parameter demanded
        
        """
        
        if name == 'P':
            return self.P[index]
        elif name == 'V':
            return self.V[index]
        elif name == 'R':
            return self.R[index]
        elif name == 'OM':
            return self.OM[index]
        elif name == 'A':
            return self.A[index]
        elif name == 'OMP':
            return self.OMP[index]
        elif name == 'J':
            return self.J[index[0],index[1]]
        else:
            print('parameter '+name+' doesn\'t exist for a sensor')
            return 0.0


class FSensor(Sensor):
    """ Class FSensor: structure to compute a 'force sensor'.
        Re-define the compute function for computing the force
        sensor kinematics using the correct index   
    """

    def compute(self):
        """ Compute the data's of the sensor by calling the symbolic function
        """
        try:
            self.computeFun(self,self.mbs,self.number+self.mbs.Nsensor)
        except TypeError:
            raise Exception('No External Force! Error FSensor.compute')
    

class Function():
    """ Class Function: contain the user and symbolic function reference
    
    :param user: dictionnary containing the user function (name, reference)
    :param sym: dictionnary containing the symbolic function (name, reference)
    :param mbs: the mbs structure of the multibody system
    
    """
    def __init__(self,mbs):
        self.user={}
        self.sym={}
        self.mbs=mbs
        self.load()
    
    def load(self):
        """Load all user and symbolic function and save their reference into the dictionnaries user
        and sym.
        If a file doesn't exist, it will ignore the problem and set the reference to 'None'
        """
#        print(sys.path)
        if self.mbs.prjpath not in sys.path:
            sys.path.append(self.mbs.prjpath)
        if self.mbs.userpath not in sys.path:
            sys.path.append(self.mbs.userpath)
        if self.mbs.sympath not in sys.path:
            sys.path.append(self.mbs.sympath)
        
#        try:
        from cons_hJ import user_cons_hJ
        self.user['cons_hJ']=user_cons_hJ
#        except ImportError:
#            self.user['cons_hJ']=None

            
        try:
            from derivatives import user_derivatives
            self.user['deriv']=user_derivatives
        except ImportError:
            self.user['deriv']=None

        
        try:
            from driven_joints import user_DrivenJoints
            self.user['djoints']=user_DrivenJoints
        except ImportError:
            self.user['djoints']=None

        
        try:
            from ext_forces import user_ExtForces
            self.user['fext']=user_ExtForces
        except ImportError:
            self.user['fext']=None

        
        try:
            from joint_forces import user_JointForces
            self.user['fjoint']=user_JointForces
        except ImportError:
            self.user['fjoint']=None

            
        try:
            from link_forces import user_LinkForces
            self.user['flink']=user_LinkForces
        except ImportError:
            self.user['flink']=None

        try:
            from link3D_forces import user_Link3DForces
            self.user['flink3D']=user_Link3DForces
        except ImportError:
            self.user['flink3D']=None

        
        try:
            from cons_jdqd import user_cons_jdqd
            self.user['jdqd']=user_cons_jdqd
        except ImportError:
            self.user['jdqd']=None

            
        
        
        #SYMBOLIC FUNCTIONS

        try:
            name='mbs_dirdyna_'+self.mbs.name
            mod=__import__(name)
            self.sym['dirdyna']=mod.dirdyna
        except ImportError:
            self.sym['dirdyna']=None

        
        try:
            name='mbs_cons_hJ_'+self.mbs.name
            mod=__import__(name)
            self.sym['cons_hJ']=mod.cons_hJ
        except ImportError:
            self.sym['cons_hJ']=None

            
        try:        
            name='mbs_cons_jdqd_'+self.mbs.name
            mod=__import__(name)
            self.sym['cons_jdqd']=mod.cons_jdqd
        except ImportError:
            self.sym['cons_jdqd']=None

        
        try:
            name='mbs_link_'+self.mbs.name
            mod=__import__(name)
            self.sym['link']=mod.link
        except ImportError:
            self.sym['link']=None

        try:
            name='mbs_link3D_'+self.mbs.name
            mod=__import__(name)
            self.sym['link3D']=mod.link3D
        except ImportError:
            self.sym['link3D']=None

        
        try:
            name='mbs_extforces_'+self.mbs.name
            mod=__import__(name)
            self.sym['extforces']=mod.extforces
        except ImportError:
            self.sym['extforces']=None

        
        try:
            name='mbs_sensor_'+self.mbs.name
            mod=__import__(name)
            self.sym['sensor']=mod.sensor
        except ImportError:
            self.sym['sensor']=None
        except AttributeError:
            self.sym['sensor']=None

        

        
        
        
        
        
