# -*- coding: utf-8 -*-

#DIRDYN.PY
#@author: Thomas Gillis
#
#Created on 3/7/2013
#Last update on 9/7/2013
#

#
# PyRobotran
# Copyright 2013, Universite catholique de Louvain, Belgium
# All rights reserved
# 
# 
# The PyRobotran software has been developed by Thomas Gillis (Louvain School of engineering)
# and Nicolas Docquier (Center for Research in Mechatronics, institute of MEchanics, Materials and 
# Civil engineering) for the Université catholique de Louvain (UCL).
#
# PyRobotran is a free software: you can redistribute it and/or modify it under the terms of 
# the BSD 3-Clause License as published below.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice, 
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice, 
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Université catholique de Louvain nor the names of
#     its contributors may be used to endorse or promote products derived from 
#     this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# For any use in commercial products or questions about licensing, you can contact
#  - the Center for Research in Mechatronics at info@cerem.be
#  - the Louvain Technology Transfer Office of UCL at ltto@uclouvain.be
#


from numpy import *
from scipy.linalg import norm




def kine_wheel(Pw,Rw,Vw,OMw,mbs):
    """ Compute the wheel characteristics
    
    Basis hypothesis: the ground is localy flat and the frame of the frozen wheel has 
    pitch and roll angles of maximum 45 degrees
    
    Convention: the axis of the wheel has to be a Y axis (R2). 
    [Rw]: wheel frame; [I]: inertial frame; [Rtg]: tangent frame; [Rsol]: frame on the ground
    aligned with the wheel
    
    
    :param Pw: absolute position vector of the center of the wheel
    :param Rw: absolute rotation matrix of the wheel
    :param Vw: absolute velocity vector of the center of the wheel
    :param OMw: absolute angular velocity of the center of the wheel
    :param mbs: the Data structure
    
    
    
    :returns rz: real radius
    :returns ancamb: camber angle
    :returns anglis: slip angle
    :returns Pct: contact point position in the inertial frame
    :returns Vct: contact point velocity in the inertial frame
    :returns Zgnd: ground level
    :returns Rsol: transformation matrix between the intertial frame and the wheel ground frame
    :returns dxF: contact point vector in the local frame
    
    """
    
#    tsim=mbs.tsim
    eyP=zeros(3+1)
    ex=zeros(3+1)
    Rsol=zeros((3+1,3+1))
    Rtg=zeros((3+1,3+1))
    
    #frame definition

    ey=array(Rw[2,:])
    eyP=array(ey)
    eyP[3]=0
    eyP=eyP/norm(eyP)
    
    
    ex=mbs_cross(eyP,array([0,0,0,1]))
    
    Rsol[1,:]=ex
    Rsol[2,:]=eyP
    Rsol[3,:]=array([0,0,0,1])
    
    Rtg[1,:]=ex
    Rtg[2,:]=ey
    Rtg[3,:]=mbs_cross(ex,ey)
    
    
    #angle cambrure
    sinf=Rw[2,3]
    ancamb=arcsin(sinf)
    
    #contact  point + velocity
    Zgnd=0.0 #USER_GROUND
    rz=(Pw[3]-Zgnd)/cos(ancamb)
    vrz=array([0,0,0,-rz])
    
        
    vrz_I=dot(Rtg.T,vrz)

    Pct=Pw+vrz_I

    
    #vitesse:

    vx1=mbs_cross(OMw,vrz_I)
    Vct = Vw + vx1;			
    Vcts = dot(Rsol,Vct)
    dxF = dot(Rw,vrz_I)
    

    Vws = dot(Rsol,Vw)
    OMw_w = dot(Rw,OMw)
    OMwf_w = array( [0,OMw_w[1], 0 , OMw_w[3]])
    OMwf = dot(Rw.T,OMwf_w)
    

    vx1 = mbs_cross(OMwf,vrz_I)

    Vct_geo = Vw + vx1			
    Vct_geo_s = dot(Rsol,Vct_geo)        
    
    if abs(Vct_geo_s[1]) >= 1e-3:
        tg_anglis = Vct_geo_s[2]/Vct_geo_s[1]
    else:
        tg_anglis = 0

    anglis = arctan(tg_anglis);


    # glisssement
    if abs(Vws[1]) >= 1e-3:
        gliss = -Vcts[1]/Vws[1]
    else:
        gliss = 0
    
    
    
    return (rz, ancamb, anglis, gliss, Pct, Vct,Zgnd, Rsol, dxF)
		
    
def mbs_cross(a,b):
    """
    Execute the cross product between a and b:  a x b
    
    :param a: numpy vector
    :param b: numpy vector
    
    a and b must have the lenght 4! The index begin at 1!
    """
    c1=a[2]*b[3]-a[3]*b[2]
    c2=a[3]*b[1]-a[1]*b[3]
    c3=a[1]*b[2]-a[2]*b[1]
    
    return array([0,c1,c2,c3])





def bakker(Frad,anglis,ancamb,gliss):
    """Computes the Bakker tyre contact model

    :param Frad: the radial force
    :param anglis: the slip angle
    :param ancamb: the roll angle
    :param gliss: the longitudinal slip ratio
    
    :return Flong: the longitudinal force
    :return Flat: the lateral force
    :return Mz: the normal force
    
    """
    Flong=0; Flat=0; Mz=0
    

    #parameters

    anglisdeg = anglis*180/pi;
    ancambdeg = ancamb*180/pi;
    
    
    FradkN = Frad/1000;
    
    k = gliss*100;
    
    C	= 1.65
    D	= -21.3*FradkN**2+1144*FradkN
    B	= (49.6*FradkN**2+226*FradkN)/(C*D*exp(0.069*FradkN))
    E	= -0.006*FradkN**2+0.056*FradkN+0.486
    phi = (1-E)*k+(E/B)*arctan(B*k) 
    
    
    Flong = D*sin(C*arctan(B*phi))
    
    
    C	= 1.3;
    D	= -22.1*FradkN**2+1011*FradkN;
    B	= (1078*sin(1.82*arctan(0.208*FradkN))*(1-0.022*abs(ancambdeg)))/(C*D);
    E	= -0.354*FradkN+0.707;
    deltaSh = 0.028*ancambdeg;
    deltaSv = (14.8*FradkN)*ancambdeg;
    phi = (1-E)*(anglisdeg+deltaSh)+(E/B)*arctan(B*(anglisdeg+deltaSh));
    Flat = -(D*sin(C*arctan(B*phi))+deltaSv);
    
    
#    Flong et Flat combinees    
    k = k/100;
    
    if (k==0 and anglis==0):
        sxs = 1;
        sys = 1;
    else:
        sxs = abs(k/(sqrt(k**2+tan(anglis)**2)));
        sys = abs(tan(anglis)/(sqrt(k**2+tan(anglis)**2)));

    
    Flong = sxs*Flong;
    Flat = sys*Flat;
    
    
#   Couple d'alignement
    C = 2.4;
    D = -2.72*FradkN**2-2.28*FradkN;
    B=(-1.86*FradkN**2-2.73*FradkN)/(C*D*exp(0.11*FradkN))*(1-0.03*abs(ancambdeg));
    E=(-0.07*FradkN**2+0.643*FradkN-4.04)*(1-0.07*abs(ancambdeg));
    deltaSh = 0.015*ancambdeg;
    deltaSv = (-0.066*FradkN**2+0.945*FradkN)*ancambdeg;
    phi = (1-E)*(anglisdeg+deltaSh)+(E/B)*arctan(B*(anglisdeg+deltaSh));
    
    Mz = D*sin(C*arctan(B*phi))+deltaSv;
    

    
    
    return (Flong,Flat,Mz)
    
    
    
    
    
    
    
    
    
    
