# -*- coding: utf-8 -*-

#XML2SYS.PY
#@author: Thomas Gillis
#         Nicolas Docquier
#Created on 2/7/2013
#Last update on 24/09/2013
#

#
# PyRobotran
# Copyright 2013, Universite catholique de Louvain, Belgium
# All rights reserved
# 
# 
# The PyRobotran software has been developed by Thomas Gillis (Louvain School of engineering)
# and Nicolas Docquier (Center for Research in Mechatronics, institute of MEchanics, Materials and 
# Civil engineering) for the Université catholique de Louvain (UCL).
#
# PyRobotran is a free software: you can redistribute it and/or modify it under the terms of 
# the BSD 3-Clause License as published below.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted
# provided that the following conditions are met:
#   * Redistributions of source code must retain the above copyright notice, 
#     this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright notice, 
#     this list of conditions and the following disclaimer in the documentation
#     and/or other materials provided with the distribution.
#   * Neither the name of the Université catholique de Louvain nor the names of
#     its contributors may be used to endorse or promote products derived from 
#     this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# For any use in commercial products or questions about licensing, you can contact
#  - the Center for Research in Mechatronics at info@cerem.be
#  - the Louvain Technology Transfer Office of UCL at ltto@uclouvain.be
#

# Modifications
#
# Date            Author        Issue
# 15/04/2014      ND            robotran/pyrobotran # 3
#



import xml.etree.ElementTree as ET
import numpy as np
import collections as col
import os

class System:
    """ Class containing the System. The system is the XML file converted into
    an object.
    
    :param name: the name of the project (string)
    :param base: the base body (object)
    :param body: dictionnary of the bodies in the project. The keys are their names and the values an object with all informations

    """
    def __init__(self,mbs):
        
        """ XML file loading """
        #chargement de l'arbre decoulant du XML        
        try:
            tree=ET.parse(mbs.mbsfile)
            root=tree.getroot()
        except IOError:
            print('ERROR: File .mbs not found!')
        
        #analyse du prelude: nom du projet
        self.name=handleName(root)
        
        # analyse de la base:
        self.base=loadBase(root)
        
        #analyse du bodytree:
            #tous les corps sont stockes sous forme d'un dictionnaire. les clés sont leurs noms et le dictionnaire contient un objet corps
        self.body=col.OrderedDict()
#        self.bodyorder=[] #liste contenant DANS L'ORDRE les corps
        for elem in root.iter('bodytree'):
            for elemBody in elem.iter('body'):
                (name,bodyobj)= loadBody(elemBody)
                self.body[name]=bodyobj
        
        #analyse des cuts
        self.cut=col.OrderedDict()
        for elem in root.iter('cuts'):
            for child in elem.iter('ball'): #on charge les balls
                (cut)=handleBall(child)
                self.cut[cut.name]=cut
            for child in elem.iter('solid'): # on charge les solids
                (cut)=handleSolid(child)
                self.cut[cut.name]=cut
            for child in elem.iter('rod'): #on charge les rods
                (cut)=handleRod(child)
                self.cut[cut.name]=cut
                
        #analyse du usermodel
        self.userModel=dicoUserModel()
        for elem in root.iter('user_model'):
            (name,model)=handleUserModel(elem)
            self.userModel.addModel(name,model)
            
        # analysis of userIO
        self.userIO = UserIOs()
        for elem in root.iter('user_variable'):
            (name,io)=handleUserIO(elem)
            setattr(self.userIO, name, io)
        
        
        #analyse des links
        self.link=col.OrderedDict()
        for elem in root.iter('link'):
            linkelem=Link()
            handlePtToPt(elem,linkelem)
            self.link[linkelem.name]=linkelem

        #analyse des link3D
        self.link3D=col.OrderedDict()
        for elem in root.iter('link3D'):
            linkelem=Link()
            handlePtToPt(elem,linkelem)
            self.link3D[linkelem.name]=linkelem

        
        #analyse des ext_forces
        
#==============================================================================
# FONCTIONS DE CHARGEMENT DES COMPOSANTS
#==============================================================================
    
def loadBase(root):
    """ Function for loading a base bloc in XML to an object Base.
    
    :param root: element of the XML tree containing a base (if possible passed through a for elem in ....iter('base') ....)
    
    :return Base object: containing the base build from the XML file
    
    """
    base= Base()
    i=0
    for elem in root.iter('base'):
        i+=1
        for child in elem.iter('gravity'):
            base.gravity= handleCoord(child)
        for child in elem.iter('point'):
            (nameP,point)=handlePoint(child)
            base.point[nameP]=point  
        
    return base
    
def loadBody(root):
    """ function for loading a body bloc in XML to a body object
    
    :param root: a XML tree containing a body (if possible passed through a for elem in ....iter('body') ...)
    
    :return (name,body) tuple containing the name and a body object: (name, body object)
    
    """
    body=Body()
    #name of the body
    for elem in root.findall('bodyname'):
        name=elem.text
        
    # parent of the body
    for elem in root.iter('parent'):
        for child in elem.iter('bodyname'):
            body.parent.append(child.text)
        for child in elem.iter('pointname'):
            body.parent.append(child.text)
    
    # joint of the body
    number=1
    for elem in root.iter('joint'):
        (namejoint,joint)=handleJoint(elem,number)
        body.joint[namejoint]=joint
        number+=1
    # body mass
    for elem in root.iter('mass'):
        body.mass=float(elem.text)
    
    #body COM
    for elem in root.iter('com'):
        for child in root.iter('coordinates'):
            body.com=handleCoord(elem)
    
    #inertia
    for elem in root.iter('inertia'):
        for child in elem.iter('Ixx'):
            body.inertia[0]=float(child.text)
        for child in elem.iter('Ixy'):
            body.inertia[1]=float(child.text)
        for child in elem.iter('Ixz'):
            body.inertia[2]=float(child.text)
        for child in elem.iter('Iyy'):
            body.inertia[3]=float(child.text)
        for child in elem.iter('Iyz'):
            body.inertia[4]=float(child.text)
        for child in elem.iter('Izz'):
            body.inertia[5]=float(child.text)
    
    #points
    for elem in root.iter('point'):
        (nameP,point)=handlePoint(elem)
        body.point[nameP]=point   

    return (name,body)

#==============================================================================
# FONCTION DE DECODAGE PROPREMENT DIT
#==============================================================================

def handleName(root):
    """ fonction which handle the transition from XML to a string for the mbsname
    
    :param root: elements of a XML tree with a name mbs
    
    :return name: name included within the mbsname tags.
    """
    result = None
    i=0
    for elem in root.iter('mbsname'):
        result=elem.text
        i+=1
    
    if i != 1: #gestion du cas où le nom est présent plusieurs fois (on garde le dernier en mémoire)
        print('ERREUR: Nombre de mbsname incorrect!')
    
    return result
    
def handleBall(root):
    """ Function to handle the ball xml element
    
    :param root: elements of a XML tree with a ball
    
    :return ball: the base object
    """
    ball=Ball()
    handlePtToPt(root,ball) #on charge les infos principales: nom, endpoint1 et endpoint2
    for elem in root.iter('ignore'): #on charge les infos secondaires
        for child in elem.iter('x'):
            ball.ignoreX=True
            ball.nbrAlgCons-=1
        for child in elem.iter('y'):
            ball.ignoreY=True
            ball.nbrAlgCons-=1
        for child in elem.iter('z'):
            ball.ignoreZ=True
            ball.nbrAlgCons-=1
    return ball
    
def handleRod(root):
    """ Function to handle the rod xml element
    
    :param root: elements of a XML tree with a rod
    
    :return rod: the rod object
    """
    rod=Rod()
    handlePtToPt(root,rod) #on charge les infos principales: nom, endpoint1 et endpoint2
    for elem in root.iter('length'):
        rod.length=float(elem.text)
    return rod

def handleSolid(root):
    """ Function to handle the solid xml element
    
    :param root: elements of a XML tree with a solid cut
    
    :return ball: the solid object
    
    """
    solid=Solid()
    handlePtToPt(root,solid) #on charge les infos principales: nom, endpoint1 et endpoint2
    return solid
    
def handlePtToPt(root,ptTopt):
    """ Function to handle the solid xml element
    
    :param root: elements of a XML tree with a solid cut
    :param ptTopt: a PtToPt object to complete
    
    """
    for elem in root.iter('name'):
        ptTopt.name=elem.text
    for elem in root.iter('endpoint1'):
        for child in elem.iter('bodyname'):
            ptTopt.endpoint1.append(child.text)
        for child in elem.iter('pointname'):
            ptTopt.endpoint1.append(child.text)
            
    for elem in root.iter('endpoint2'):
        for child in elem.iter('bodyname'):
            ptTopt.endpoint2.append(child.text)
        for child in elem.iter('pointname'):
            ptTopt.endpoint2.append(child.text)
    
    
    
def handleCoord(root):
    """ function which handle the transition from XML to a vector for coordinates
    
    :param root: element of a tree with coordinates (if possible passed through a for elem in ....iter('coordinates')...)
    
    :return coord: vector with the x,y and z coordinates: [X,Y,Z]
    
    """
    coord=np.zeros(3)
    for elem in root.iter('x'):
        coord[0]=float(elem.text)
    for elem in root.iter('y'):
        coord[1]=float(elem.text)
    for elem in root.iter('z'):
        coord[2]=float(elem.text)
    return coord

def handlePoint(root):
    """ function which handle the transition from XML to a tuple containing the
    
    name and a vector with the position of the point
    :param root: element of a tree with point (if possible passed through a for elem in ....iter('point')...)
    
    :return (name,point): a tuple with the name and vector with the x,y and z coordinates of the point and a boolean force which indicate if the point support some external forces: (name,[X,Y,Z],force)
    
    """
    point=Point()
    for elem in root.iter('pointname'):
        name=elem.text
    for elem in root.iter('coordinates'):
        point.value=handleCoord(elem)
    for elem in root.iter('extforce'):
        point.force = True
        point.forceName=elem.attrib['name']
    for elem in root.iter('sensor'):
        point.sensor=True
        point.sensorName=elem.attrib['name']
    
    return (name,point)

def handleJoint(root,index):
    """ function which handle the transcription from a XML file to a joint.
    
    :param root: element of a XML tree with joint (if possible passed through a for elem in ....iter('joint') ....)
    :param index: the index of the joint
    
    :return (name,joint): name of the joint and the joint object
    
    """
    joint=Joint(index)
    for elem in root.iter('jointname'):
        name=elem.text
    for elem in root.iter('type'):
        joint.type=elem.text
    for elem in root.iter('nature'):
        joint.nature=elem.text
    for elem in root.iter('initialvalue'):
        for child in elem.iter('q'):
            joint.initialval[0]=float(child.text)
        for child in elem.iter('qd'):
            joint.initialval[1]=float(child.text)
        for child in elem.iter('qdd'):
            joint.initialval[2]=float(child.text)
    
    return (name,joint)
    
def handleUserModel(root):
    """ function which handle the transcription from a XML file to a user model.
    
    :param root: element of a XML tree with user model
    
    :return (name,usr): name of the user model and the userModel object
    
    """
    usr=UserModel()
    for elem in root.findall('name'):
        name=elem.text
    for elem in root.iter('parameter'):
        for child in elem.iter('name'):
            nameParam=child.text
        param=Parametre()
        for child in elem.iter('type'): #pour les scalaires:
            param.type=child.text
        for child in elem.iter('value'):
            param.value.append(float(child.text))
        param.value=np.array(param.value)
        usr.dicoParam[nameParam]=param
                
    return (name,usr)


def handleUserIO(root):
    """ function which handle the transcription from a XML file to a user io.
    
    :param root: element of a XML tree with user io (<user_variable>)
    
    :return (name,usr): name of the io and the userIO object
    
    """

    
    for elem in root.findall('varname'):
        name=elem.text
    for elem in root.findall('type'):
        typ=elem.text
    for elem in root.findall('size'):
        size=int(elem.text)

    usr=UserIO(name, typ, size)
                    
    return (name,usr)
            
        
    
#==============================================================================
#     CLASSES DEFINIES POUR CONTENIR LES ELEMENTS
#==============================================================================
    
class Base:
    """ Classe of the base:
        
    :param gravity: vector containing the gravity
    :param point: dictionnary containing the point. The keys are the name and the values a vector containing the position.
        
    """
    def __init__(self):
        self.gravity=np.zeros(3)
        self.point=col.OrderedDict() #dictionnaire reprenant les points d'ancrages: leurs petit nom et leurs coordonnées sous forme d'un vecteur.
        
class Body:
    """ Classe des corps:
        
    :param parent: list of parents: [parent body, parent point]
    :param joint: dictionnary containing joints. The keys are the name of the joint and the values are joint object.
    :param mass: body mass (number)
    :param com: centre of mass (vector)
    :param inertia: inertia matrice (vector)
    :param point: dictionnary of the point. The keys are the point name and the values are vectors containing position.
        
    """
    def __init__(self):
        self.parent=[] #liste des noms des parents[corps parent, point parent]
        self.joint=col.OrderedDict() #dictionnaire contenant par nom les joints
        self.mass=0
        self.com=np.zeros(3)
        self.inertia=np.zeros(6)
        self.point=col.OrderedDict() #dictionnaire contenant les points
        

class Joint:
    """ Classe des joints:
        
    :param type: kind of the joint (string): R2, T3, R1,...
    :param nature: nature of the joint (string): dependant, driven, independant
    :param initialval: the initial values of the joint given by user (array): [q0, qd0, qdd0]
    :param nbr: number of the joint in the list for one body form 1 to...we count joints from the body parent to the current body. The number one is the joint attached to the parent body.
        
    """
    def __init__(self,number): #numero du joint voir doc class Joint
        self.nbr=number
        self.type=''
        self.nature=''
        self.initialval=np.zeros(3) #valeurs initiales [q0, qd0, qdd0]

class PtToPtElement:
    """ Class of the elments that exists between two points
    
    :param name: the name of the elemnt
    :param endpoint1: the point 1
    :param endpoint2: the point 2
    
    """
    def __init__(self):
        self.name='' #nom de la cut
        self.endpoint1=[] #liste des noms des points de depart
        self.endpoint2=[] #liste des noms des points d'arrivee

class Link(PtToPtElement):
    """Class of the link (extends PtToPtElement)
    
    """
    def __init__(self):
        PtToPtElement.__init__(self)
        

class Cut(PtToPtElement):
    """ Class of the cut (extends PtToPtElement)
    
    :param nbrAlgCons: the number of algebraic constraints
    :param type: the type of the cut (solid, ball,,rod)
    
    """
    def __init__(self,nbrAlgCons=0,typeCut='solid'):
        PtToPtElement.__init__(self)
        self.nbrAlgCons=nbrAlgCons
        self.type=typeCut #solid, ball ou rod en fonction du type de cut

class Ball(Cut):
    """ Class for the ball cut (extends Cut)
    
    :param ignoreX: boolean. True if we have to ignore the X constraint
    :param ignoreY: boolean. True if we have to ignore the Y constraint
    :param ignoreZ: boolean. True if we have to ignore the Z constraint
    
    """
    def __init__(self):
        Cut.__init__(self,3,'ball')
        self.ignoreX=False
        self.ignoreY=False
        self.ignoreZ=False

class Solid(Cut):
    """ Class for the solid cut (extends cut)
    """
    def __init__(self):
        Cut.__init__(self,6,'solid')

class Rod(Cut):
    """ Class for the rod cut (extends cut)
    
    :param length: the length of the rod
    
    """
    def __init__(self):
        Cut.__init__(self,1,'rod')
        self.length=1

class Point:
    """ Class of the points
    :param value: the position of the point
    :param force: True if the point has a external force
    :param forceName: the name of the force
    :param sensor: True if the point has a sensor
    :param sensorName: The name of the sensor
    
    """
    def __init__(self):
        self.value=np.zeros(3)
        self.force=False
        self.forceName=''
        self.sensor=False
        self.sensorName=''

class UserModel:
    """ Class of the user models
    :param Nx: the number of state variable in the userModel
    :param dicoParam: the dictionnary containing the name of the parameter and the corresponding object
    
    """
    def __init__(self):
        self.Nx=0 #nombre de variable d'etat reliees au user_model
        self.dicoParam={}
        
    def __getattr__(self,param):
        return self.dicoParam[param].value

    def __dir__(self): 
        """ Override the __dir__ method to ensure that the tab completion (when using 
            environment like ipython) works well.
        """
        return self.dicoParam.keys()




    # add the following lines for making it possible to save the object using PIckle
    def __getstate__(self):
        return self.__dict__
    def __setstate__(self, d):
        self.__dict__.update(d)


        
class Parametre:
    """ Class of the parameters of an user model
    
    :param type: the kind of the parameter
    :param value: the value of this parameter
    """
    
    def __init__(self,kind='scalar'):
        self.type=kind
        self.value=[]

        

class dicoUserModel:
    """Class dicoUserModel
    Used to store the userModels
    
    :param dico: dictionnary containing the userModels (nameModel, userModelObject)    
    
    """
    
    def __init__(self):
        self.dico={}
    
    def addModel(self,name,model):
        """ Add the model to the dictionnary
        
        :param name: string containing the name of the model
        :param model: userModel object
        
        """
        self.dico[name]=model
    
    def __getattr__(self,modelname):
        if modelname in self.dico:
            return self.dico[modelname]
       # else:
       #     super(Object, self).__getattr__(self,modelname)
        
        
    def __dir__(self): 
        """ Override the __dir__ method to ensure that the tab completion (when using 
            environment like ipython) works well.
        """
        return self.dico.keys()    


    # add the following lines for making it possible to save the object using PIckle
    # but is is not enought!!!!!
    def __getstate__(self):
        return self.__dict__
    def __setstate__(self, d):
        self.__dict__.update(d)

class UserIO:
    """ Class UserIO
    Used to store the info about one user variable (Input/Output)
    """
    
    def __init__(self, name, typ, size):
        self.name = name
        self.type = typ
        self.size = size  

    def __str__(self):
        return 'UserIO - name: ' +self.name+' - type: '+self.type+' - size: '+str(self.size)

        
    def __repr__(self):
        return 'UserIO:\nname: ' +self.name+'\ntype: '+self.type+'\nsize: '+str(self.size)

    
class UserIOs( object ):
    """ Class UserIOs
    Used to store UserIO obects
    """
    
    def getUserIONames(self):
        """ return a list containing the names of the UserIO
        contained in this container
        """
        return self.__dict__.keys()
    
    def __setattr__(self, name, val):
        if(name in self.getUserIONames()):
            # if the IO already exists, modify the value
            print("In progress...Should be modified for modifying the value of the UserIO")
        else:
            # if the IO does not exists, add it
            super(UserIOs, self).__setattr__(name, val)



    # add the following lines for making it possible to save the object using PIckle        
    def __getstate__(self):
        return self.__dict__
    def __setstate__(self, d):
        self.__dict__.update(d)
